#!/usr/bin/env node
import { __assign, __rest, __read, __spreadArray } from 'tslib';
import fs from 'fs';
import commander from 'commander';
import chalk from 'chalk';
import Ajv from 'ajv';
import 'ts-array-extensions';
import path from 'path';
import yaml from 'yaml';
import prettier from 'prettier';

/* eslint-disable no-console */
var o = process.stdout;
var e = process.stdout;
var progressLogger = function (start) {
    var last = 0;
    return function (m) {
        o.cursorTo(start);
        o.write(m);
        if (m.length < last) {
            o.write(' '.repeat(last - m.length));
        }
        last = m.length;
    };
};
var success = function () { return chalk.green(' -> SUCCESS'); };
var warn = function (message) {
    return chalk.yellow(" -> WARNING: " + message);
};
var progress = function (message) {
    return chalk.blue(" -> " + message);
};
var processLogger = function () {
    var log = function (m) {
        o.write("\n" + m);
    };
    return {
        error: function (m) {
            e.write("\n" + chalk.red(m));
        },
        log: log,
        create: function (m) {
            log(m);
            var l = progressLogger(m.length);
            l(progress('...'));
            return l;
        }
    };
};

var mapProperties = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema = function (schema) {
    /**
     * This currently does nothing, but is a placeholder for removing
     * unsupported features from Open Api
     */
    var properties = schema.properties, rest = __rest(schema, ["properties"]);
    return __assign(__assign({}, rest), (properties ? { properties: mapProperties(properties) } : undefined));
};

var dereference = function (references, $ref, prefix) {
    if (!$ref.startsWith(prefix)) {
        throw new Error("Invalid $ref " + $ref);
    }
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    var maybe = references === null || references === void 0 ? void 0 : references[name];
    if (!maybe)
        throw new Error("Failed to deference " + $ref);
    return maybe;
};

var dereferenceParameter = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.parameters, $ref, '#/parameters/');
};

var dereferenceResponseObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference(doc.responses, $ref, '#/responses/');
};

function isReferenceObject(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var methodsOfPathItemObject = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' || method === 'parameters') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        return methodsOfPathItemObject(pathObject).map(function (_a) {
            var _b = __read(_a, 2), method = _b[0], operationObject = _b[1];
            if (!('operationId' in operationObject) ||
                !operationObject.operationId) {
                throw new Error("OperationId required for " + method + " " + path);
            }
            var operationId = operationObject.operationId, _c = operationObject.parameters, parameters = _c === void 0 ? [] : _c, responses = operationObject.responses;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject(p) ? dereferenceParameter(p, document) : p;
            });
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject(response)
                        ? dereferenceResponseObject(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                responses: inlinedResponses
            };
        });
    });
};

var initUpper = function (value) {
    return value.replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};

var parametersTypeNameTemplate = function (operationId, iin) { return initUpper(operationId) + "Request" + initUpper(iin); };

var requestValidatorTemplate = function (types, jsonSchemaFilename) { return "\nimport Ajv, { ValidateFunction } from 'ajv';\nimport { RequestHandler, Request } from 'express';\nimport schema from './" + jsonSchemaFilename + "';\n\nconst ajv = new Ajv({ strict: false });\najv.addSchema(schema);\n\ntype V = [ValidateFunction, (req: Request) => unknown];\n\nconst validator = (validations: V[]): RequestHandler => {\n  return (req, res, next) => {\n    if (validations.every(([fn, sel]) => fn(sel(req)))) {\n      return next();\n    }\n\n    res.status(400).send({\n      errors: validations\n        .flatMap(([v]) => v.errors)\n        .map(e => e.message ?? '') ?? ['Unknown validation error']\n    });\n  };\n};\n\n" + types
    .map(function (_a) {
    var operationId = _a.operationId, properties = _a.properties;
    return "\nexport const " + operationId + " = validator([\n  " + properties
        .map(function (_a) {
        var requestPropertyName = _a.requestPropertyName, typeName = _a.typeName;
        return "[ajv.getSchema('#/definitions/" + typeName + "')!, p => p." + requestPropertyName + "]";
    })
        .join(',\n') + "\n]);";
})
    .join('\n') + "\n"; };

var responseBodyTypeNameTemplate = function (operationId, statusCode) { return "" + initUpper(operationId) + statusCode + "ResponseBody"; };

var responseTypeNameTemplate = function (operationId) {
    return initUpper(operationId) + "Response";
};

var responseTypesTemplate = function (schemaTypesModuleName, types) { return "\nimport * as Schema from './" + schemaTypesModuleName + "';\n\nexport type ResponseWithData<TStatus, TData> = {\n  status: TStatus;\n  body: TData;\n};\n\n" + types
    .map(function (_a) {
    var type = _a.type, childTypes = _a.childTypes;
    return "\nexport type " + type + " = " + childTypes
        .map(function (_a) {
        var statusCode = _a.statusCode, bodyType = _a.bodyType;
        return "ResponseWithData<" + statusCode + ", " + (bodyType === 'unknown' ? bodyType : "Schema." + bodyType) + ">";
    })
        .join(' | ') + "\n\n";
})
    .join('') + "\n"; };

var fromV2$6 = function (document, log) {
    var definitions = {};
    mapOperations(document).forEach(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        parameters
            .groupBy(function (p) { return p.in; })
            .forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!['body', 'path', 'query', 'header', 'formData'].includes(key)) {
                throw new Error("Unsupported parameter in " + key);
            }
            if (key === 'body') {
                if (values.length > 1) {
                    throw new Error('Only a single body parameter per operation is supported');
                }
                var schema_1 = values[0].schema;
                var typeName_1 = parametersTypeNameTemplate(operationId, 'body');
                log === null || log === void 0 ? void 0 : log(progress("adding " + typeName_1));
                definitions[typeName_1] = convertSchemaToJsonSchema(schema_1);
                return;
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = __rest(v, ["in", "name", "required"]);
                schema.properties[name] = rest;
                if (required) {
                    schema.required = __spreadArray(__spreadArray([], __read(schema.required)), [name]);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding " + typeName));
            definitions[typeName] = convertSchemaToJsonSchema(schema);
        });
        responses.forEach(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(schema);
        });
    });
    Object.keys(document.definitions).forEach(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding " + key));
            definitions[key] = convertSchemaToJsonSchema(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};
var fromV3$6 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateJsonSchema = function (_a) {
    var openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Generating JSON schema from Open API Document");
    var schema = 'swagger' in openApiDocument ? fromV2$6(openApiDocument, log) : fromV3$6();
    log === null || log === void 0 ? void 0 : log(progress("Validating JSON schema"));
    var ajv = new Ajv({ strict: false });
    ajv.addSchema(schema);
    Object.keys(schema.definitions).forEach(function (key) {
        log === null || log === void 0 ? void 0 : log(progress("Validating " + key));
        ajv.getSchema("#/definitions/" + key);
    });
    log === null || log === void 0 ? void 0 : log(success());
    return schema;
};

var fromV2$5 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses;
        var bodyArg = parameters.first(function (p) { return p.in === 'body'; });
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var responseTypeName = responseTypeNameTemplate(operationId);
        var interpolationPath = '`${this.baseUrl}' +
            path.replace(/\{(?:.*?)\}/g, function (x) { return '$' + x; }) +
            (queryArgs.any() ? '?${query}' : '') +
            '`';
        var argsType = __spreadArray(__spreadArray(__spreadArray(__spreadArray([], __read((pathArgs.any()
            ? ["Schema." + parametersTypeNameTemplate(operationId, 'path')]
            : []))), __read((queryArgs.any()
            ? ["Schema." + parametersTypeNameTemplate(operationId, 'query')]
            : []))), __read((headerArgs.any()
            ? ["Schema." + parametersTypeNameTemplate(operationId, 'header')]
            : []))), __read((bodyArg
            ? [
                "{ body: Schema." + parametersTypeNameTemplate(operationId, 'body') + " }"
            ]
            : []))).join('&');
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var fetchOpts = "{\n        " + __spreadArray(__spreadArray([
            'method'
        ], __read((headerArgs.any()
            ? ["headers: {" + headerArgs.map(function (h) { return h.name; }).join(',') + "}"]
            : []))), __read((bodyArg ? ['body: JSON.stringify(body)'] : []))).join(',\n') + "\n      }";
        responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var hasResponseBody = responses.some(function (r) { return r.response.schema; });
        var paramNames = parameters
            .filter(function (p) { return p.in !== 'body'; })
            .map(function (p) { return p.name; })
            .concat(bodyArg ? ['body'] : []);
        return "\npublic async " + operationId + "(" + (argsType ? "args: " + argsType : '') + "): Promise<Responses." + responseTypeName + "> {\n  " + (argsType ? "const {" + paramNames.join(', ') + "} = args;" : '') + "\n\n  " + (queryArgs.any()
            ? "const query = qs.stringify({ " + queryArgs
                .map(function (q) { return q.name; })
                .join(', ') + " });"
            : '') + "\n\n  const method = '" + method.toUpperCase() + "',\n  const url = " + interpolationPath + ";\n\n  const response = await fetch(url, " + fetchOpts + ");\n\n  const { status } = response;\n\n  " + (hasResponseBody
            ? 'const responseBody = (await response.json()) as unknown;'
            : '') + "\n\n  this.logger?.('REST API Call', {\n    " + __spreadArray(__spreadArray(__spreadArray(__spreadArray([
            'method',
            'url'
        ], __read((bodyArg ? ['requestBody: body'] : []))), [
            'status'
        ]), __read((hasResponseBody ? ['responseBody'] : []))), __read(headerArgs.concat(pathArgs).map(function (a) { return a.name; }))).join(',\n') + "\n  });\n\n  return {\n    status,\n    body: " + (hasResponseBody ? 'responseBody' : 'undefined') + "\n  } as Responses." + responseTypeName + ";\n}";
    });
};
var fromV3$5 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateClient = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, clientResponseTypesModuleName = _a.clientResponseTypesModuleName, schemaTypesModuleName = _a.schemaTypesModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating client');
    var types = 'swagger' in openApiDocument ? fromV2$5(openApiDocument, log) : fromV3$5();
    var code = "\nimport qs from 'qs';\nimport { fetch } from 'cross-fetch';\nimport * as Responses from './" + clientResponseTypesModuleName + "';\nimport * as Schema from './" + schemaTypesModuleName + "';\n  \ntype LogFn = (message: string, data: Record<string, unknown>) => void;\n\nexport class Client {\n  public constructor(\n    public readonly baseUrl: string,\n    private readonly logger?: LogFn\n  ) {}\n  \n  " + types.join('\n\n') + "\n}\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$4 = function (document) {
    return mapOperations(document).map(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        var childTypes = responses.map(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            var bodyType = schema
                ? responseBodyTypeNameTemplate(operationId, statusCode)
                : 'unknown';
            return {
                statusCode: statusCode === 'default' ? 'number' : statusCode,
                bodyType: bodyType
            };
        });
        var type = responseTypeNameTemplate(operationId);
        return {
            type: type,
            childTypes: childTypes
        };
    });
};
var fromV3$4 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateClientResponseTypes = function (_a) {
    var openApiDocument = _a.openApiDocument, logger = _a.logger, schemaTypesModuleName = _a.schemaTypesModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating response types');
    var types = 'swagger' in openApiDocument ? fromV2$4(openApiDocument) : fromV3$4();
    var code = responseTypesTemplate(schemaTypesModuleName, types);
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var generateIndex = function (_a) {
    var logger = _a.logger, clientModuleName = _a.clientModuleName, output = _a.output, routerModuleName = _a.routerModuleName, schemaTypesModuleName = _a.schemaTypesModuleName, serverStubsModuleName = _a.serverStubsModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating index');
    var code = "\nexport * from './" + schemaTypesModuleName + "';\n" + (output.includes('CLIENT') ? "export * from './" + clientModuleName + "';" : '') + "\n" + (output.includes('STUBS') ? "export * from './" + serverStubsModuleName + "';" : '') + "\n" + (output.includes('SERVER') ? "export * from './" + routerModuleName + "';" : '') + "\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var initLower = function (value) {
    return value.replace(/(?:^|\s)[A-Z]/g, function (m) { return m.toLowerCase(); });
};

var fromV2$3 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            return "Schema." + responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var bodyParamType = parameters.some(function (p) { return p.in === 'body'; })
            ? "Schema." + parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var queryParamsType = parameters.some(function (p) { return p.in === 'query'; })
            ? "Schema." + parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var pathParamsType = parameters.some(function (p) { return p.in === 'path'; })
            ? "Schema." + parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var typeName = initUpper(operationId) + "RequestHandler";
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return {
            code: "export type " + typeName + " =\n          RequestHandler<\n            " + pathParamsType + ",\n            " + responseBodyType + ",\n            " + bodyParamType + ",\n            " + queryParamsType + ",\n            Record<string, any>,\n            " + statusCodes + ">",
            operationId: operationId
        };
    });
};
var fromV3$3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRequestHandlersTypes = function (_a) {
    var schemaTypesModuleName = _a.schemaTypesModuleName, logger = _a.logger, openApiDocument = _a.openApiDocument;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types for request handlers');
    var types = 'swagger' in openApiDocument ? fromV2$3(openApiDocument, log) : fromV3$3();
    var code = "\n  import { RequestHandler } from '@laurence79/express-async-request-handler';\n  import * as Schema from './" + schemaTypesModuleName + "';\n  \n  " + types.map(function (t) { return t.code; }).join('\n\n') + "\n\n  export type RequestHandlers = {\n    " + types
        .map(function (t) {
        return initLower(t.operationId) + ": " + initUpper(t.operationId) + "RequestHandler";
    })
        .join(';\n') + "\n  };\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$2 = function (document) {
    return mapOperations(document).compactMap(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters;
        return {
            operationId: operationId,
            properties: parameters
                .map(function (p) { return p.in; })
                .distinct()
                .map(function (p) { return ({
                requestPropertyName: p === 'header' ? 'headers' : p,
                typeName: parametersTypeNameTemplate(operationId, p)
            }); })
        };
    });
};
var fromV3$2 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRequestValidators = function (_a) {
    var jsonSchemaFilename = _a.jsonSchemaFilename, openApiDocument = _a.openApiDocument, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating request body validators');
    var types = 'swagger' in openApiDocument ? fromV2$2(openApiDocument) : fromV3$2();
    var code = requestValidatorTemplate(types, jsonSchemaFilename);
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var fromV2$1 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, operationId = _a.operationId;
        var handlerType = "RequestHandlers." + initUpper(operationId) + "RequestHandler";
        var handlerInstanceName = "" + initLower(operationId);
        var requestHandlersPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + requestHandlersPath));
        return {
            route: "." + method + "('" + requestHandlersPath + "', Validators." + operationId + ", " + handlerInstanceName + ")",
            handlerInstanceName: handlerInstanceName,
            handlerType: handlerType
        };
    });
};
var fromV3$1 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateRouter = function (_a) {
    var requestSchemaValidatorsModuleName = _a.requestSchemaValidatorsModuleName, logger = _a.logger, openApiDocument = _a.openApiDocument, requestHandlersModuleName = _a.requestHandlersModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating router');
    var types = 'swagger' in openApiDocument ? fromV2$1(openApiDocument, log) : fromV3$1();
    var argumentList = types
        .map(function (_a) {
        var handlerInstanceName = _a.handlerInstanceName;
        return handlerInstanceName;
    })
        .join(', ');
    var code = "\n  import { Router } from 'express';\n  import * as RequestHandlers from './" + requestHandlersModuleName + "';\n  import * as Validators from './" + requestSchemaValidatorsModuleName + "';\n  \n  export const router = ({ " + argumentList + " }: RequestHandlers.RequestHandlers): Router => {\n    return Router()\n      " + types.map(function (t) { return t.route; }).join('\n') + "\n  }\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

/* eslint-disable prefer-template */
var fromV2 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding " + method + " " + path));
        var bodyArg = parameters.first(function (p) { return p.in === 'body'; });
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? "Schema." + parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? "Schema." + parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? "Schema." + parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = bodyArg
            ? "Schema." + parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            return "Schema." + responseBodyTypeNameTemplate(operationId, statusCode);
        })
            .join(' | ') || 'unknown';
        var statusCodes = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var methodStubType = "MethodStub<\n        " + headersType + ",\n        " + pathParamsType + ",\n        " + queryParamsType + ",\n        " + requestBodyType + ",\n        " + statusCodes + ",\n        " + responseBodyType + "\n      >";
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":" + x.substr(1, x.length - 2); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};
var fromV3 = function () {
    throw new Error('OpenApi 3 not supported');
};
var generateServerStub = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, schemaTypesModuleName = _a.schemaTypesModuleName;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server stubs');
    var types = 'swagger' in openApiDocument ? fromV2(openApiDocument, log) : fromV3();
    var code = "\nimport { RequestHandler } from '@laurence79/express-async-request-handler';\nimport { Router } from 'express';\nimport * as Schema from './" + schemaTypesModuleName + "';\n  \ntype RequestLog<THeaders, TParams, TQuery, TBody> = {\n  headers: THeaders;\n  params: TParams;\n  query: TQuery;\n  body: TBody;\n};\n\ntype ResponseStub<TStatus extends number, TBody> = {\n  statusCode: TStatus;\n  body: TBody;\n};\n\ntype MethodStub<\n  THeaders,\n  TParams,\n  TQuery,\n  TRequestBody,\n  TStatus extends number,\n  TResponseBody\n> = {\n  requestHandler: RequestHandler<\n    TParams,\n    TResponseBody,\n    TRequestBody,\n    TQuery,\n    Record<string, unknown>,\n    TStatus\n  >;\n  respondWith(data: ResponseStub<TStatus, TResponseBody>): void;\n  callLog(): RequestLog<THeaders, TParams, TQuery, TRequestBody>[];\n};\n\nconst methodStub = <\n  THeaders,\n  TParams,\n  TQuery,\n  TRequestBody,\n  TStatus extends number,\n  TResponseBody\n>(): MethodStub<\n  THeaders,\n  TParams,\n  TQuery,\n  TRequestBody,\n  TStatus,\n  TResponseBody\n> => {\n  const log: RequestLog<THeaders, TParams, TQuery, TRequestBody>[] = [];\n\n  let responseStub: ResponseStub<TStatus, TResponseBody> | null = null;\n\n  return {\n    requestHandler: (req, res) => {\n      const { headers, params, query, body } = req;\n\n      log.push({\n        headers: (headers as unknown) as THeaders,\n        params,\n        query,\n        body\n      });\n\n      if (!responseStub) {\n        throw new Error('Method not stubbed');\n      }\n\n      res.status(responseStub.statusCode).send(responseStub.body);\n    },\n\n    respondWith(data) {\n      responseStub = data;\n    },\n\n    callLog() {\n      return log;\n    }\n  };\n};\n\ntype MethodStubs = {\n  " + types
        .map(function (_a) {
        var operationId = _a.operationId, methodStubType = _a.methodStubType;
        return "readonly " + operationId + ": " + methodStubType + ";";
    })
        .join('\n') + "\n};\n\nexport type ServiceStub = {\n  readonly middleware: RequestHandler;\n  readonly reset: () => void;\n  readonly stubs: () => MethodStubs;\n};\n\nexport const serviceStub = (): ServiceStub => {\n  let methodStubs: MethodStubs;\n  let currentRouter: RequestHandler;\n\n  const reset = () => {\n    methodStubs = {\n      " + types
        .map(function (_a) {
        var operationId = _a.operationId;
        return operationId + ": methodStub()";
    })
        .join(',\n') + "\n    };\n\n    currentRouter = Router()\n      " + types
        .map(function (_a) {
        var method = _a.method, operationId = _a.operationId, requestHandlerPath = _a.requestHandlerPath;
        return "." + method + "('" + requestHandlerPath + "', methodStubs." + operationId + ".requestHandler)";
    })
        .join('\n') + "\n  };\n\n  reset();\n\n  return {\n    reset,\n    middleware: (...args: Parameters<RequestHandler>) => currentRouter(...args),\n    stubs: () => methodStubs\n  };\n};\n";
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var typeDefForObject = function (objectSchema, recursiveLookup) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject);
        return "" + propertyName + (isRequired ? '' : '?') + ": " + propertyType;
    });
    return "{ " + propertyList.join('; ') + " }";
};

var typeDefForEnum = function (members, onwardLookup) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"" + e + "\"";
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'object') {
            if (Array.isArray(e)) {
                throw new Error("Don't know what to do with an array-like enum member");
            }
            return onwardLookup(e);
        }
        return e.toString();
    })
        .join(' | ');
};

var typeDefForReference = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return name;
};

var typeDefForSchema = function (schema) {
    var _a;
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if ('$ref' in schema && schema.$ref) {
        return typeDefForReference(schema.$ref);
    }
    if (!('type' in schema)) {
        throw new Error("Schema nodes must have a type.");
    }
    var type = schema.type, enumProp = schema.enum;
    if (enumProp) {
        return typeDefForEnum(enumProp, typeDefForSchema);
    }
    if (type === 'object') {
        return typeDefForObject(schema, typeDefForSchema);
    }
    if (type === 'boolean' ||
        type === 'null' ||
        type === 'number' ||
        type === 'string') {
        return type;
    }
    if (type === 'integer') {
        return 'number';
    }
    if (type === 'array') {
        var items = schema.items;
        if (!items) {
            throw new Error("Array nodes must have an items property.");
        }
        if (Array.isArray(items)) {
            throw new Error("Array items property must be singular.");
        }
        return "Array<" + typeDefForSchema(items) + ">";
    }
    throw new Error("Unrecognized type " + ((_a = type === null || type === void 0 ? void 0 : type.toString()) !== null && _a !== void 0 ? _a : ''));
};

var generateTypesFromJsonSchema = function (_a) {
    var jsonSchema = _a.jsonSchema, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating typescript types from JSON schema');
    var definitions = jsonSchema.definitions;
    if (!definitions)
        return '';
    var types = Object.keys(definitions).compactMap(function (typeName) {
        if (typeof typeName !== 'string')
            return undefined;
        var definition = definitions[typeName];
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "export type " + typeName + " = " + typeDefForSchema(definition) + ";";
    });
    var code = types.join('\n\n');
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var loadOpenApiDocument = function (_a) {
    var openApiDocumentFilename = _a.openApiDocumentFilename, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create("Loading swagger file " + openApiDocumentFilename);
    if (!fs.existsSync(openApiDocumentFilename)) {
        throw new Error("File " + openApiDocumentFilename + " does not exist");
    }
    var extension = path.extname(openApiDocumentFilename);
    var data = fs.readFileSync(openApiDocumentFilename, 'UTF-8');
    var doc = (function () {
        if (['.yaml', '.yml'].includes(extension)) {
            return yaml.parse(data);
        }
        if (extension === '.json') {
            return JSON.parse(data);
        }
        throw new Error("Unsupported extension " + extension);
    })();
    log === null || log === void 0 ? void 0 : log(success());
    return doc;
};

var prettierStyle = {
    arrowParens: 'avoid',
    semi: true,
    trailingComma: 'none',
    singleQuote: true,
    printWidth: 80
};
var writeFiles = function (_a) {
    var files = _a.files, outputDirectory = _a.outputDirectory, logger = _a.logger;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Writing files');
    log === null || log === void 0 ? void 0 : log(progress('Creating output directory'));
    fs.mkdirSync(outputDirectory, { recursive: true });
    files.forEach(function (_a) {
        var filename = _a.filename, content = _a.content;
        log === null || log === void 0 ? void 0 : log(progress(filename));
        var extension = path.extname(filename);
        var parser = (function () {
            switch (extension) {
                case '.ts':
                    return 'typescript';
                case '.json':
                    return 'json';
                default:
                    return undefined;
            }
        })();
        var fullFileName = path.join(outputDirectory, filename);
        var formattedContent = (function () {
            try {
                return prettier.format(content, __assign(__assign({}, prettierStyle), { parser: parser }));
            }
            catch (e) {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
                log === null || log === void 0 ? void 0 : log(warn(e.message));
                return content;
            }
        })();
        fs.writeFileSync(fullFileName, formattedContent);
    });
    log === null || log === void 0 ? void 0 : log(success());
};

var defaultOptions = {
    clientModuleName: 'client',
    clientResponseTypesModuleName: 'clientResponseTypes',
    jsonSchemaFilename: 'schema.json',
    output: ['CLIENT', 'SERVER', 'STUBS'],
    outputDirectory: 'generated',
    requestHandlersModuleName: 'handlers',
    requestSchemaValidatorsModuleName: 'requestValidators',
    routerModuleName: 'router',
    schemaTypesModuleName: 'schemaTypes',
    serverStubsModuleName: 'serverStubs'
};
var generateCode = function (inputOptions) {
    var options = __assign(__assign({}, defaultOptions), inputOptions);
    var openApiDocument = loadOpenApiDocument(options);
    var jsonSchema = generateJsonSchema(__assign(__assign({}, options), { openApiDocument: openApiDocument }));
    var renderDeps = __assign(__assign({}, options), { openApiDocument: openApiDocument,
        jsonSchema: jsonSchema });
    var when = function (targets, generator) {
        if (targets.union(options.output).any()) {
            return generator();
        }
        return null;
    };
    var files = [
        when(['SERVER'], function () { return ({
            filename: options.jsonSchemaFilename,
            content: JSON.stringify(jsonSchema)
        }); }),
        {
            filename: options.schemaTypesModuleName + ".ts",
            content: generateTypesFromJsonSchema(renderDeps)
        },
        when(['SERVER'], function () { return ({
            filename: options.requestSchemaValidatorsModuleName + ".ts",
            content: generateRequestValidators(renderDeps)
        }); }),
        when(['SERVER'], function () { return ({
            filename: options.requestHandlersModuleName + ".ts",
            content: generateRequestHandlersTypes(renderDeps)
        }); }),
        when(['SERVER'], function () { return ({
            filename: options.routerModuleName + ".ts",
            content: generateRouter(renderDeps)
        }); }),
        when(['CLIENT'], function () { return ({
            filename: options.clientResponseTypesModuleName + ".ts",
            content: generateClientResponseTypes(renderDeps)
        }); }),
        when(['CLIENT'], function () { return ({
            filename: options.clientModuleName + ".ts",
            content: generateClient(renderDeps)
        }); }),
        when(['STUBS'], function () { return ({
            filename: options.serverStubsModuleName + ".ts",
            content: generateServerStub(renderDeps)
        }); }),
        {
            filename: 'index.ts',
            content: generateIndex(renderDeps)
        }
    ].compact();
    writeFiles(__assign(__assign({}, options), { files: files }));
};

var packageJsonContent = fs.readFileSync(__dirname + "/../package.json", 'utf-8');
var packageJson = JSON.parse(packageJsonContent);
var version = packageJson.version;
var cmd = new commander.Command('generate')
    .description('Generates Express types and functions from the spec.')
    .arguments('<spec-filename> <output-directory>')
    .option('--client', 'output client modules')
    .option('--server', 'output server modules')
    .option('--stubs', 'output server stub modules')
    .action(function (specFilename, outputDirectory, options) {
    var logger = processLogger();
    try {
        generateCode({
            openApiDocumentFilename: specFilename,
            output: __spreadArray(__spreadArray(__spreadArray([], __read((options.client ? ['CLIENT'] : []))), __read((options.server ? ['SERVER'] : []))), __read((options.stubs ? ['STUBS'] : []))),
            outputDirectory: outputDirectory,
            logger: logger
        });
        logger.log(chalk.green('\nCompleted successfully!\n\n'));
    }
    catch (e) {
        logger.error("\nFAILED. " + (e instanceof Error ? e.message : 'Unknown error') + "\n\n");
    }
});
commander.program.version(version).addCommand(cmd).parse();

export { generateCode };
